@@ -2,11 +2,22 @@
 
 import numpy as np
 from astropy.constants import c
+from pyuvdata import UVBeam
 from scipy.interpolate import RectBivariateSpline
 from typing import Optional, Sequence
 
 from . import conversions
 
+# This enables us to put in profile decorators that will be no-ops if no profiling
+# library is being used.
+try:
+    profile
+except NameError:
+
+    def profile(fnc):
+        """No-op profiling decorator."""
+        return fnc
+
 
 def construct_pixel_beam_spline(bm_cube):
     """Construct bivariate spline for pixelated beams for all antennas.
@@ -32,12 +43,8 @@ def construct_pixel_beam_spline(bm_cube):
             "imaginary components separately."
         )
 
-    if len(bm_cube.shape) == 5:
-        # Polarized beam
-        nax, nfeed, nant, bm_pix, _ = bm_cube.shape
-    else:
-        nax = nfeed = 1
-        nant, bm_pix, _ = bm_cube.shape
+    # Polarized beam
+    nax, nfeed, nbeam, bm_pix, _ = bm_cube.shape
 
     # x and y coordinates of beam
     bm_pix_x = np.linspace(-1, 1, bm_pix)
@@ -51,8 +58,8 @@ def construct_pixel_beam_spline(bm_cube):
         for p2 in range(nfeed):
             spl_feeds = []
 
-            # Loop over antennas
-            for i in range(nant):
+            # Loop over beams/antennas
+            for i in range(nbeam):
                 # Linear interpolation of primary beam pattern.
                 spl = RectBivariateSpline(
                     bm_pix_y, bm_pix_x, bm_cube[p1, p2, i], kx=1, ky=1
@@ -63,6 +70,7 @@ def construct_pixel_beam_spline(bm_cube):
     return splines
 
 
+@profile
 def vis_cpu(
     antpos: np.ndarray,
     freq: float,
@@ -70,9 +78,10 @@ def vis_cpu(
     crd_eq: np.ndarray,
     I_sky: np.ndarray,
     bm_cube: Optional[np.ndarray] = None,
-    beam_list: Optional[Sequence[np.ndarray]] = None,
+    beam_list: Optional[Sequence[UVBeam]] = None,
     precision: int = 1,
     polarized: bool = False,
+    beam_idx: Optional[np.ndarray] = None,
 ):
     """
     Calculate visibility from an input intensity map and beam model.
@@ -105,9 +114,10 @@ def vis_cpu(
         Shape=(NSRCS,).
     bm_cube : array_like, optional
         Pixelized beam maps for each antenna. If ``polarized=False``,
-        shape=``(NANT, BM_PIX, BM_PIX)``, otherwise
-        shape=``(NAX, NFEED, NANT, BM_PIX, BM_PIX)``. Only one of ``bm_cube`` and
-        ``beam_list`` should be provided.
+        shape=``(NBEAMS, BM_PIX, BM_PIX)``, otherwise
+        shape=``(NAX, NFEED, NBEAMS, BM_PIX, BM_PIX)``. Only one of ``bm_cube`` and
+        ``beam_list`` should be provided. If NBEAMS != NANT, then `beam_idx` must be
+        provided also.
     beam_list : list of UVBeam, optional
         If specified, evaluate primary beam values directly using UVBeam
         objects instead of using pixelized beam maps. Only one of ``bm_cube`` and
@@ -123,6 +133,10 @@ def vis_cpu(
         Whether to simulate a full polarized response in terms of nn, ne, en,
         ee visibilities. See Eq. 6 of Kohn+ (arXiv:1802.04151) for notation.
         Default: False.
+    beam_idx
+        Optional length-NANT array specifying a beam index for each antenna.
+        By default, either a single beam is assumed to apply to all antennas or
+        each antenna gets its own beam.
 
     Returns
     -------
@@ -160,39 +174,58 @@ def vis_cpu(
         I_sky.ndim == 1 and I_sky.shape[0] == nsrcs
     ), "I_sky must have shape (NSRCS,)."
 
+    # Get the number of unique beams
+    if bm_cube is not None:
+        nbeam = bm_cube.shape[2 if polarized else 0]
+    else:
+        nbeam = len(beam_list)
+
+    # Check the beam indices
+    if beam_idx is None:
+        if nbeam == 1:
+            beam_idx = np.zeros(nant, dtype=int)
+        elif nbeam == nant:
+            beam_idx = np.arange(nant, dtype=int)
+        else:
+            raise ValueError(
+                "If number of beams provided is not 1 or nant, beam_idx must be provided."
+            )
+    else:
+        assert beam_idx.shape == (nant,), "beam_idx must be length nant"
+        assert all(
+            0 <= i < nbeam for i in beam_idx
+        ), "beam_idx contains indices greater than the number of beams"
+
     if beam_list is None:
         bm_pix = bm_cube.shape[-1]
         complex_bm_cube = np.any(np.iscomplex(bm_cube))
         if polarized:
-            assert bm_cube.shape == (nax, nfeed, nant, bm_pix, bm_pix), (
-                "bm_cube must have shape (NAXES, NFEEDS, NANTS, BM_PIX, BM_PIX) if "
+            assert bm_cube.shape == (nax, nfeed, nbeam, bm_pix, bm_pix), (
+                "bm_cube must have shape (NAXES, NFEEDS, NBEAMS, BM_PIX, BM_PIX) if "
                 f"polarized=True. Shape wanted: {(nax, nfeed, nant, bm_pix, bm_pix)}; "
                 f"shape given: {bm_cube.shape}"
             )
-        elif bm_cube.shape != (1, 1, nant, bm_pix, bm_pix):
-            assert bm_cube.shape == (nant, bm_pix, bm_pix), (
-                "bm_cube must have shape (NANTS, BM_PIX, BM_PIX) "
-                "or (1, 1, nant, bm_pix, bm_pix) if polarized=False. "
-                f"Shape wanted: {(nant, bm_pix, bm_pix)}; "
+        elif bm_cube.shape != (1, 1, nbeam, bm_pix, bm_pix):
+            assert bm_cube.shape == (nbeam, bm_pix, bm_pix), (
+                "bm_cube must have shape (NBEAMS, BM_PIX, BM_PIX) "
+                "or (1, 1, nbeam, bm_pix, bm_pix) if polarized=False. "
+                f"Shape wanted: {(nbeam, bm_pix, bm_pix)}; "
                 f"shape given: {bm_cube.shape}"
             )
             bm_cube = bm_cube[np.newaxis, np.newaxis]
-    else:
-        if polarized and any(b.beam_type != "efield" for b in beam_list):
-            raise ValueError("beam type must be efield if using polarized=True")
-        elif not polarized and any(
-            (
-                b.beam_type != "power"
-                or getattr(b, "Npols", 1) > 1
-                or b.polarization_array[0] not in [-5, -6]
-            )
-            for b in beam_list
-        ):
-            raise ValueError(
-                "beam type must be power and have only one pol (either xx or yy) if polarized=False"
-            )
-
-        assert len(beam_list) == nant, "beam_list must have length nant"
+    elif polarized and any(b.beam_type != "efield" for b in beam_list):
+        raise ValueError("beam type must be efield if using polarized=True")
+    elif not polarized and any(
+        (
+            b.beam_type != "power"
+            or getattr(b, "Npols", 1) > 1
+            or b.polarization_array[0] not in [-5, -6]
+        )
+        for b in beam_list
+    ):
+        raise ValueError(
+            "beam type must be power and have only one pol (either xx or yy) if polarized=False"
+        )
 
     # Intensity distribution (sqrt) and antenna positions. Does not support
     # negative sky. Factor of 0.5 accounts for splitting Stokes I between
@@ -203,10 +236,7 @@ def vis_cpu(
     ang_freq = 2.0 * np.pi * freq
 
     # Zero arrays: beam pattern, visibilities, delays, complex voltages
-    A_s = np.zeros((nax, nfeed, nant, nsrcs), dtype=complex_dtype)
     vis = np.zeros((nax, nfeed, ntimes, nant, nant), dtype=complex_dtype)
-    tau = np.zeros((nant, nsrcs), dtype=real_dtype)
-    v = np.zeros((nant, nsrcs), dtype=complex_dtype)
     crd_eq = crd_eq.astype(real_dtype)
 
     # Precompute splines using pixelized beams
@@ -215,52 +245,63 @@ def vis_cpu(
         if complex_bm_cube:
             splines_im = construct_pixel_beam_spline(bm_cube.imag)
 
+    im = 0
+
     # Loop over time samples
     for t, eq2top in enumerate(eq2tops.astype(real_dtype)):
         # Dot product converts ECI cosines (i.e. from RA and Dec) into ENU
         # (topocentric) cosines, with (tx, ty, tz) = (e, n, u) components
         # relative to the center of the array
         tx, ty, tz = crd_top = np.dot(eq2top, crd_eq)
+        above_horizon = tz > 0
+        tx = tx[above_horizon]
+        ty = ty[above_horizon]
+        nsrcs_up = len(tx)
+
+        A_s = np.zeros((nax, nfeed, nbeam, nsrcs_up), dtype=complex_dtype)
+        tau = np.zeros((nant, nsrcs_up), dtype=real_dtype)
+        v = np.zeros((nant, nsrcs_up), dtype=complex_dtype)
 
         # Primary beam response
         if beam_list is None:
             # Primary beam pattern using pixelized primary beam
-            for i in range(nant):
+            for i in range(nbeam):
                 # Extract requested polarizations
                 for p1 in range(nax):
                     for p2 in range(nfeed):
                         # The beam pixel grid has been reshaped in the order
                         # ty,tx, which implies m,l order
-                        A_s[p1, p2, i] = splines_re[p1][p2][i](ty, tx, grid=False)
+
+                        re = splines_re[p1][p2][i](ty, tx, grid=False)
+
                         if complex_bm_cube:
-                            A_s[p1, p2, i] += 1.0j * splines_im[p1][p2][i](
-                                ty, tx, grid=False
-                            )
+                            im = 1.0j * splines_im[p1][p2][i](ty, tx, grid=False)
+
+                        A_s[p1, p2, i] = re + im
         else:
 
             # Primary beam pattern using direct interpolation of UVBeam object
             az, za = conversions.enu_to_az_za(enu_e=tx, enu_n=ty, orientation="uvbeam")
-            for i in range(nant):
+            for i in range(nbeam):
                 interp_beam = beam_list[i].interp(
                     az_array=az, za_array=za, freq_array=np.atleast_1d(freq)
                 )[0]
 
                 if polarized:
-                    A_s[:, :, i] = interp_beam[:, 0, :, 0, :]  # spw=0 and freq=0
+                    interp_beam = interp_beam[:, 0, :, 0, :]
                 else:
                     # Here we have already asserted that the beam is a power beam and
                     # has only one polarization, so we just evaluate that one.
-                    A_s[:, :, i] = np.sqrt(interp_beam[0, 0, 0, :, :])
+                    interp_beam = np.sqrt(interp_beam[0, 0, 0, 0, :])
 
-        # Horizon cut
-        A_s = np.where(tz > 0, A_s, 0)
+                    A_s[:, :, i] = interp_beam
 
         # Check for invalid beam values
         if np.any(np.isinf(A_s)) or np.any(np.isnan(A_s)):
             raise ValueError("Beam interpolation resulted in an invalid value")
 
         # Calculate delays, where tau = (b * s) / c
-        np.dot(antpos, crd_top, out=tau)
+        np.dot(antpos, crd_top[:, above_horizon], out=tau)
         tau /= c.value
 
         # Component of complex phase factor for one antenna
@@ -269,17 +310,15 @@ def vis_cpu(
         np.exp(1.0j * (ang_freq * tau), out=v)
 
         # Complex voltages.
-        v *= Isqrt
+        v *= Isqrt[above_horizon]
 
         # Compute visibilities using product of complex voltages (upper triangle).
         # Input arrays have shape (Nax, Nfeed, [Nants], Nsrcs
+        v = A_s[:, :, beam_idx] * v[np.newaxis, np.newaxis, :]
+
         for i in range(len(antpos)):
             vis[:, :, t, i : i + 1, i:] = np.einsum(
-                "ijln,jkmn->iklm",
-                A_s[:, :, i : i + 1].conj()
-                * v[np.newaxis, np.newaxis, i : i + 1].conj(),
-                A_s[:, :, i:] * v[np.newaxis, np.newaxis, i:],
-                optimize=True,
+                "ijln,jkmn->iklm", v[:, :, i : i + 1].conj(), v[:, :, i:], optimize=True
             )
 
     # Return visibilities with or without multiple polarization channels
